// ------------------------------------
// Version 1.0 : Algorithmes de graphes
// ------------------------------------

// Copyright (C) 2003 (Lacomme, Prins, Sevaux)

// Ce programme (ainsi que la bibliothque objets livre avec le livre)
// est libre, vous pouvez le redistribuer et/ou
// le modifier selon les termes de la
// Licence Publique Gnrale GNU publie par la
// Free Software Foundation (version 2 ou bien toute autre version
// ultrieure choisie par vous).
// Ce programme est distribu car potentiellement utile,
// mais SANS AUCUNE GARANTIE, ni explicite ni implicite,
// y compris les garanties de commercialisation ou
// d'adaptation dans un but spcifique. Reportez-vous  la
// Licence Publique Gnrale GNU pour plus de dtails.
// Vous devez avoir reu une copie de la Licence
// Publique Gnrale GNU en mme temps que
// ce programme ; si ce n'est pas le cas, crivez  la
// Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
// 02111-1307, tats-Unis.

// Diffrents sites donnent des copies officielles ou non de cette licence :
// http://www.linux-france.org/article/these/gpl.html
// http://www.gnu.org/copyleft/gpl.html


unit Unit6;

interface

uses
  SysUtils, Types, Classes, Graphics, Controls, Forms, Dialogs,
  StdCtrls, ExtCtrls;

type
  TForm6 = class(TForm)
    RadioGroup1: TRadioGroup;
    RadioButton1: TRadioButton;
    RadioButton2: TRadioButton;
    RadioButton3: TRadioButton;
    RadioButton4: TRadioButton;
    RadioButton5: TRadioButton;
    CheckBox1: TCheckBox;
    Button1: TButton;
    RadioButton6: TRadioButton;
    RadioButton7: TRadioButton;
    procedure Button1Click(Sender: TObject);
  private
    { Dclarations prives }
  public
    { Dclarations publiques }
  end;

var
  Form6: TForm6;

implementation

{$R *.dfm}

uses U_TYPE_GRAPHES, U_BASE_GRAPHES, U_GRAPHES,Unit1;

procedure TForm6.Button1Click(Sender: TObject);
CONST

Seed   = 197211719;   {Germe du generateur aleatoire}
NTest  =       020;   {Nb de graphes a tester}
NNodes =       800;   {Taille du graphe matriciel}
T0     =    2500.0;   {Recuit: temperature initiale}
Coef   =    0.9900;   {Recuit: coef de refroidissement}
Eps    =      0.01;   {Recuit: temperature d'arret}
NPass  =        16;   {Recuit: nombre de passes}
NT     =        50;   {Tabou : taille de la liste taboue (max 50)}
NItMax =       100;   {Tabou : nb max d'iterations}
NItWI  =       100;   {Tabou : nb max d'iterations sans amelioration}

VAR

M      : T_GRAPHE_MATRICIEL;      {Graphe matriciel}
Cycle  : TNodeInfo;   {Cycle trouve}
C      : Cost;        {Son cout}
Choice : Integer;     {N de l'heuristique choisie}
Rep    : Char;        {Reponse a la question sur type de graphe}
TDeb   : TDateTime;   {Heure de debut}
D      : TDateTime;   {Duree}
Test   : Integer;     {N de test}
NIter  : LongInt;     {Nb d'iterations (tabou et recuit)}
SavSeed: LongInt;     {Sauvegarde du germe (pour Recuit)}
SumC   : LongInt;     {Somme des couts sur la serie de graphes}
SumD   : TDateTime;   {Somme des durees}

//--------------------------------------------
// Ok: verifie un cycle Cycle et son cout Cost
//--------------------------------------------

Function OK (var Cycle:TNodeInfo; C:Cost): Boolean;
Var Flag: TNodeBool;
    CC  : Cost;
    i,j : Node;
Begin
   With M do begin
      Ok := True;
      CC := 0;
      For i := 1 to p_NX do Flag[i] := False;
      For i := 1 to p_NX do begin
         j := Cycle[i];
         If Flag[j] then OK := False else Flag[j] := True;
         CC := CC + p_A[j,Cycle[i+1]]
      End;
      If CC <> C then Ok := False
   End
End;

//--------------------------------------------
// RandEuclid: genere aleatoirement un graphe euclidien (carre de cote 707)
//--------------------------------------------

Procedure RandEuclid (var M:T_GRAPHE_MATRICIEL);
Var X,Y: Array[Node] of Real;
    i,j: Node;
Begin
   With M do begin
      p_NX    := NNodes;
      p_NY    := 0;
      p_Simple:= True;
      p_NoArc := 0;
      
      For i := 1 to p_NX do begin
         p_A[i,i] := 0;
         X[i] := 707.0*Random; {Pour avoir des distances entre 0 et 1000}
         Y[i] := 707.0*Random
      End;
      For i := 1 to p_NX do For j := i+1 to p_NX do begin
         p_A[i,j] := Round (Sqrt(Sqr(X[i]-X[j])+Sqr(Y[i]-Y[j])));
         p_A[j,i] := p_A[i,j]
      End
   End
End;

Begin

   M:=T_GRAPHE_MATRICIEL.Create;

   If RadioButton1.checked=True Then
     Choice:=1
   else
     If RadioButton2.checked=True Then
       Choice:=2
     else
       If RadioButton3.checked=True Then
         Choice:=3
       else
         If RadioButton4.checked=True Then
          Choice:=4
         else
           If RadioButton5.checked=True Then
             Choice:=5
           else
             If RadioButton6.checked=True Then
               Choice:=6
             else
               Choice:=7;

   RandSeed := Seed;
   Form1.Memo1.Lines.Add ('EVALUATION DES HEURISTIQUES POUR LE PVC');
   If CheckBox1.Checked=True Then
     REP:='O'
   else
     REP:='N';

   SumC := 0;
   SumD := 0.0;

   For Test := 1 to NTest do begin
     // GotoXY (1,WhereY-1);
      Form1.Memo1.Lines.Add  ('Test n'+IntToStr(Test)+'. ');
      If Rep = 'N'
      Then M.RandMatrix  (NNodes,0,Simple,LoopLess,not Layered,
                        1,999,0,1.0,RandSeed,False)
      Else RandEuclid (M);
      TDeb := M.Chrono;
      Case Choice of
         1: M.NearestNeib (Cycle,C);
         2: M.NearestIns  (Cycle,C);
         3: M.FarthestIns (Cycle,C);
         4: M.BestIns     (Cycle,C);
         5: M.TwoOpt      (Cycle,C);
         6: Begin
               SavSeed := RandSeed;
               M.SimAn (T0,Coef,Eps,NPass,Seed,False,Cycle,C,NIter);
               RandSeed := SavSeed
            End;
         7: M.Tabu (NT,NItMax,NItWI,Cycle,C,NIter)
      End;
      D := M.Chrono-TDeb;
      Form1.Memo1.Lines.Add ('Cout:'+IntToStr(C)+', Duree: '+FormatDateTime('hh:mm:ss:zzz',D)+', NIter: '+IntToStr(NIter));
      If not Ok(Cycle,C) then M.Error ('Resultats invalides!');
      SumC := SumC + C;
      SumD := SumD + D
   End;
   Form1.Memo1.Lines.Add('');
   Form1.Memo1.Lines.Add ('Serie de '+IntToStr(NTest)+' problemes:');
   Form1.Memo1.Lines.Add ('Cout moyen   : '+FloatToStr(SumC/NTest));
   Form1.Memo1.Lines.Add ('Duree moyenne: '+FormatDateTime('hh:mm:ss:zzz',(SumD/NTest)));

   M.Pause   ('Click pour continuer...');
   M.Destroy;
   Form6.Close;
End;


end.
